class MqttConfigableItem():
    topic = ''
    type = ''
    value = ''

    def __init__(self,topic,value,type='string'):
        self.topic = topic
        self.type = type
        self.value = value

class helper:

    def update_from_topic(self, mqtt_cofig_obj, topic, value, space_len=0):
        target_type_name = 'MqttConfigableItem'
        if space_len / 8 >= 3:
            return

        for this_item in dir(mqtt_cofig_obj):
            if this_item[:1] != '_':
                attr = getattr(mqtt_cofig_obj,this_item)
                type_name =type(attr).__name__
                # space = ' ' * space_len

                if type_name == target_type_name:
                    # For better understanding, we rename attr.
                    configable_item = attr  
                    # print ('aaaa', space + configable_item, type_name)
                    for type_value_topic in dir(configable_item):
                        # print('bbbb',type_value_topic)
                        if type_value_topic == 'topic':
                            topic_string = getattr(configable_item,type_value_topic)
                            # print('cccc', type_value_topic,topic_string)
                            if topic_string == topic:
                                # print('ffff',type_value_topic,topic_string)
                                if topic_string == topic:
                                    # print('RRRRRRRRRRRR', configable_item, type_value_topic, value)
                                    #TODO: type checking here.
                                    setattr(configable_item,'value',value)
                else:
                    self.find_member(attr, target_type_name, space_len + 4)

    '''
    Valuable function, might be useful in future.
    '''
    def find_member(self, var, target_type_name ='MqttConfigableItem', space_len=0 ):
        if space_len / 8 >= 3:
            return
        for this_item in dir(var):
            if this_item[:1] != '_':
                attr = getattr(var,this_item)
                type_name =type(attr).__name__
                space = ' ' * space_len
                # print( space + '---------------------')
                # print (space + this_item, type_name)

                if type_name == target_type_name:
                    # For better understanding, we rename attr.
                    target_obj = attr  
                    print (space + this_item, type_name)
                    for attr_name in dir(target_obj):
                        if attr_name[:1] != '_':
                            print ('     --', attr_name, getattr(target_obj, attr_name)) 
                else:
                    self.find_member(attr, target_type_name, space_len + 4)




if __name__ == '__main__':
    class mqtt_config_test:
        right = MqttConfigableItem('gobot/test/right',1)
        left = MqttConfigableItem('gobot/test/right',2)
        hello = MqttConfigableItem('gobot/test/hello','Hello World')

    hh = helper()
    # hh.find_member(mqtt_config_test)
    hh.update_from_topic(mqtt_config_test,'gobot/test/hello', 'aaaabbbb')
    print (mqtt_config_test.hello.value)


from von.mqtt_agent import g_mqtt, g_mqtt_broker_config


class RemoteVar_mqtt():
    def __init__(self, mqtt_topic: str, default_value):
        self.__mqtt_topic = mqtt_topic
        self.__value = default_value
        # self.__default_value = default_value
        # self.remote_value = None
        # self.local_value = default_value
        # self.auto_copy_to_local = False
        # self.auto_copy_to_remote = False
        # self.__on_copy_to_local_callback = None
        g_mqtt.subscribe(self.__mqtt_topic)
        g_mqtt.append_on_received_message_callback(self.__on_mqtt_agent_received_message)

    def set(self, new_value):
        if new_value != self.__value:
            self.__value = new_value
            g_mqtt.publish(self.__mqtt_topic, new_value)

    def get(self):
        return self.__value

    # def set_callback_on_sync_to_local(self, callback):
    #     '''
    #     This is optional , if the app want to get local update informing immediately.
    #     '''
    #     self.__on_copy_to_local_callback = callback
    
    def __on_mqtt_agent_received_message(self, mqtt_message_topic, mqtt_message_payload):
        if mqtt_message_topic == self.__mqtt_topic:
            self.__value = mqtt_message_payload

    # def Copy_LocalToRemote(self):
    #     if self.local_value != self.remote_value:
    #         g_mqtt.publish(self.__mqtt_topic, self.local_value)

    # def Copy_RemoteToLocal(self):
    #     self.local_value = self.remote_value
    #     if self.__on_copy_to_local_callback != None:
    #         self.__on_copy_to_local_callback()
    


if __name__ == "__main__":

    g_mqtt.connect_to_broker(g_mqtt_broker_config)
    while not g_mqtt.paho_mqtt_client.is_connected():
        pass
    print("connected to mqtt broker.......")


    test_id = 3
    if test_id == 1:
        # put this line to anywhere.
        g_mqtt.publish('test/auto_sync/age', 6)
        print("check publishing result with  MQTT.fx Ver1.7.1 ,  Quit this with Ctrl+C")
        while True:
            pass
    
    if test_id == 2:
        var_hello =  RemoteVar_mqtt(mqtt_topic='test/auto_sync/hello', default_value='hello')
        
        # print ('default value=', var_hello.default_value)
        print ('remote value=', var_hello.remote_value)
        var_hello.Copy_LocalToRemote()
        var_hello.auto_copy_to_local=True
        print("Instruction 1/2: With any MQTT client, publish a message :  topic='test/auto_sync/hello', payload='aaabbb'")
        while var_hello.local_value == 'hello':
            pass
        print (var_hello.local_value)
        print("Instruction 2/2: Watch this topic,  Run this tester again, should see the payload becomes default value")

    if test_id == 3:
        ir_state = RemoteVar_mqtt(mqtt_topic="twh/" + str(221109) + "/ir_state",  default_value="unknown")
        while True:
            print (ir_state.remote_value)
        



import paho.mqtt.client as mqtt
import cv2
import logging

from von.singleton import Singleton
from von.terminal_font import TerminalFont


class MQTT_BrokerConfig:
    broker = 'voicevon.vicp.io'
    port = 1883
    uid = 'von'
    password = 'von1970'
    client_id = 'test_22111815'

class MqttAgent_ReceivedMessage():
    def __init__(self, topic, payload):
        self.topic = topic
        self.payload = payload
        self.updated = True
    
class MqttAgent_ReceivedDiction():
    def __init__(self) -> None:
        self.__newest_message = [MqttAgent_ReceivedMessage('','')]

    def FindItem(self, topic) -> MqttAgent_ReceivedMessage:
        for message in self.__newest_message:
            if message.topic == topic:
                return message
        return None

    def OnReceivedMessage(self, topic, payload):
        item = self.FindItem(topic)
        if item is None:
            new_message = MqttAgent_ReceivedMessage(topic, payload)
            self.__newest_message.append(new_message)
        else:
            item.payload = payload
            item.updated = True

    def has_updated_payload(self, topic) -> bool:
        item = self.FindItem(topic)
        return item.updated

    def FetchPayload(self, topic):
        item = self.FindItem(topic)
        if item is None:
            return None
        else:
            item.updated = False
            return item.payload


class MqttAgent(metaclass=Singleton):
    '''
    When received a message, There are two ways to get it:
    1.  set a __on_message_callbacks().   Note:  Not in main thread.
    2.  MqttAgent().RxBuffer.FetchPayload() 
    '''
    def __init__(self):
        self.paho_mqtt_client = None

        self.__YELLOW = TerminalFont.Color.Fore.yellow
        self.__GREEN = TerminalFont.Color.Fore.green
        self.__RED = TerminalFont.Color.Fore.red
        self.__RESET = TerminalFont.Color.Control.reset
        self.__on_message_callbacks = []
        self.__do_debug_print_out = False
        self.RxBuffer = MqttAgent_ReceivedDiction()

        

    def _on_phao_mqtt_client_connected_callback(self, client, userdata, flags, rc):
        if rc == 0:
            print(self.__GREEN + "MQTT connected OK. Start subscribe.  Returned code=" + self.__RESET,rc)
            # self.auto_subscribe()
        else:
            print("Bad connection Returned code= ", rc)      

    def __on_received_mqtt_message(self, client, userdata, message):
        '''
        When mqtt client received a message, this agent will callback many functions those care about the message.
        '''
        if self.__do_debug_print_out:
            #print("MQTT message received ", str(message.payload.decode("utf-8")))
            print("MQTT message topic=", message.topic)
            print('MQTT message payload=', message.payload)
            print("MQTT message qos=", message.qos)
            print("MQTT message retain flag=", message.retain)
        payload = str(message.payload.decode("utf-8"))
        for invoking in self.__on_message_callbacks:
            invoking(message.topic, payload)
        self.RxBuffer.OnReceivedMessage(message.topic, message.payload)

    def connect_to_broker(self, config: MQTT_BrokerConfig) -> mqtt.Client:
        self.paho_mqtt_client = mqtt.Client(config.client_id)  # create new instance
        self.paho_mqtt_client.on_connect = self._on_phao_mqtt_client_connected_callback     # binding call back function 
        self.paho_mqtt_client.on_message = self.__on_received_mqtt_message
        self.paho_mqtt_client.username_pw_set(username=config.uid, password=config.password)
        self.paho_mqtt_client.connect(config.broker, port=config.port)
        if self.paho_mqtt_client.is_connected():
            print(self.__GREEN + '[Info]: MQTT has connected to: %s' % config.broker + self.__RESET)
        else:
            print(self.__RED + '[Warn]: MQTT has NOT!  connected to: %s, Is trying auto connect backgroundly.' % config.broker + self.__RESET)

        #self.paho_mqtt_client.loop_forever()
        self.paho_mqtt_client.loop_start()
        # self.paho_mqtt_client.loop_stop()
        return self.paho_mqtt_client

    def append_on_received_message_callback(self, callback, do_debug_print_out=False):
        '''
        will call back on received any message.
        Says not invoved to topic. 
        '''
        self.__on_message_callbacks.append(callback)
        self.__do_debug_print_out = do_debug_print_out

    def subscribe(self, topic, qos=0):
        self.paho_mqtt_client.subscribe(topic, qos)
        self.RxBuffer.OnReceivedMessage(topic, "")


    def publish_cv_image(self, topic, cv_image, retain=True):
      # return image as mqtt message payload
        is_success, img_encode = cv2.imencode(".jpg", cv_image)
        if is_success:
            img_pub = img_encode.tobytes()
            self.paho_mqtt_client.publish(topic, img_pub, retain=retain)

    def publish_file_image(self, topic, file_name, retain=True):
        with open(file_name, 'rb') as f:
            byte_im = f.read()
        self.paho_mqtt_client.publish('sower/img/bin',byte_im )

    def publish(self, topic, payload):
        self.paho_mqtt_client.publish(topic, payload, qos=2, retain =True)
    

g_mqtt_broker_config = MQTT_BrokerConfig()
g_mqtt = MqttAgent()

if __name__ == "__main__":
    test_topic = "twh/221109/gcode_feed"
    g_mqtt.connect_to_broker(g_mqtt_broker_config)
    g_mqtt.subscribe(test_topic)
    count = 0
    while True:
        rx = g_mqtt.RxBuffer
        if rx.has_updated_payload(test_topic):
            payload = rx.FetchPayload(test_topic)
            print(count, payload)
            count += 1
        







# https://www.devdungeon.com/content/colorize-terminal-output-python
# https://www.geeksforgeeks.org/print-colors-python-terminal/

class TerminalFont:
    class Color:
        class Control:  
            '''
            Full name: Perfect_color_text
            '''
            reset='\033[0m'
            bold='\033[01m'
            disable='\033[02m'
            underline='\033[04m'
            reverse='\033[07m'
            strikethrough='\033[09m'
            invisible='\033[08m'

        class Fore:
            '''
            Full name: Perfect_fore_color
            ''' 
            black='\033[30m'
            red='\033[31m'
            green='\033[32m'
            orange='\033[33m'
            blue='\033[34m'
            purple='\033[35m'
            cyan='\033[36m'
            lightgrey='\033[37m'
            darkgrey='\033[90m'
            lightred='\033[91m'
            lightgreen='\033[92m'
            yellow='\033[93m'
            lightblue='\033[94m'
            pink='\033[95m'
            lightcyan='\033[96m'

        class Background: 
            '''
            Full name: Perfect_background_color
            ''' 
            black='\033[40m'
            red='\033[41m'
            green='\033[42m'
            orange='\033[43m'
            blue='\033[44m'
            purple='\033[45m'
            cyan='\033[46m'
            lightgrey='\033[47m'

class CvColor:
    line = (0,255,0)  
    circle = (255,255,0)

if __name__ == "__main__":
    print (TerminalFont.Color.Fore.yellow + 'Hello world.   ' +TerminalFont.Color.Fore.red + 'Hey!')


from von.amq_agent import  g_amq, g_amq_broker_config
from von.mqtt_agent import g_mqtt,g_mqtt_broker_config

class AmqMqtt_Bridge():
    def __init__(self, queue_name:str, mqtt_topic:str) -> None:
        g_amq.Subscribe(queue_name=queue_name)
        g_amq.process_data_events()
        self.amq_queue_name = queue_name
        self.amq_payload = g_amq.fetch_message_payload(self.amq_queue_name)

        self.mqtt_topic_feed = mqtt_topic
        g_mqtt.publish(self.mqtt_topic_feed, self.amq_payload)

        self.mqtt_topic_feedback = mqtt_topic + '/fb'
        g_mqtt.subscribe(self.mqtt_topic_feedback)
        # self.mqtt_feedback_message = None

    def spin_once(self):
        feedback_payload = g_mqtt.RxBuffer.FetchPayload(self.mqtt_topic_feedback)
        if feedback_payload == self.amq_payload:
            # feed a new message out, This message is come from amq
            new_payload = g_amq.fetch_message_payload(self.amq_queue_name)
            if new_payload is not None:
                print("AmqMqtt_Bridge::spin_once()  topic= " + self.mqtt_topic_feed + "   payload= " + new_payload.decode('utf-8'))
                g_mqtt.publish(self.mqtt_topic_feed, new_payload)
                self.amq_payload = new_payload


class AmqMqtt_Bridges():
    def __init__(self) -> None:
        self.all_bridges = []

    def Append(self, amq_queue_name:str):
        '''
        amq_queue_name should not contain '_'
        '''
        mqtt_publish_topic = amq_queue_name.replace("_", "/")
        new_bridge = AmqMqtt_Bridge(amq_queue_name, mqtt_publish_topic)
        self.all_bridges.append(new_bridge)

    def spin_once(self):
        for bridge in self.all_bridges:
            bridge.spin_once()

if __name__ == '__main__':
    g_amq.connect_to_broker(g_amq_broker_config)
    g_mqtt.connect_to_broker(g_mqtt_broker_config)    
    while not g_mqtt.paho_mqtt_client.is_connected():
        pass
    print("connected to mqtt broker.......")

    a = AmqMqtt_Bridges()
    a.Append('twh_221109_gcode')
    while True:
        g_amq.process_data_events(0.1)
        a.spin_once()

    # s =  AmqMqtt_Bridge('twh_221109_gcode', 'twh/221109/gcode_feed')
    # while True:
    #     g_amq.process_data_events()
    #     s.spin_once()


import serial.tools.list_ports

class DevicesHelper():
    def __init__(self):
        pass

    def serial_port_from_chip_name(self, chip_name):
        myports = [tuple(p) for p in list(serial.tools.list_ports.comports())]
        for port_name,chip,detail in myports:
            if chip == chip_name:
                return port_name

    def serial_port_from_location(self, location):
        myports = [tuple(p) for p in list(serial.tools.list_ports.comports())]
        for port_name,chip,detail in myports:
            xx =  detail[-5:]
            print(xx)
            if xx == location:
                return port_name

    def serial_port_list_all(self):
        myports = [tuple(p) for p in list(serial.tools.list_ports.comports())]
        for port_name,chip,detail in myports:
            print('-------------------------------------')
            print('name =', port_name)
            print('chip=',chip)
            print('detail=',detail)

   

    
if __name__ == "__main__":
    helper = DevicesHelper()
    helper.serial_port_list_all()
    portname = helper.serial_port_from_location('1-2.4')
    print(portname)
    portname = helper.serial_port_from_location('1-2.3')
    print(portname)


